{+--------------------------------------------------------------------------+
 | Class:       TLineStream
 | Created:     8.97 -4.98
 | Author:      Martin Waldenburg
 | Copyright    1997, all rights reserved.
 | Description: TLineStream gives a buffered access to the lines of a file.
 |              Don't forget to flush the Memory after writing.
 | Version:     1.8
 | State:       FreeWare
 | Disclaimer:
 | This is provided as is, expressly without a warranty of any kind.
 | You use it at your own risc.
 +--------------------------------------------------------------------------+}

unit mwLineStream;

interface

uses
  SysUtils, Classes;

type
  TLineStream = class(TFileStream)
  private
    fMaxMemorySize: Longint;
    fMemorySize: LongInt;
    fMemory: PChar;
    fLineStart: LongInt;
    fMemoryPos: LongInt;
    fEof: Boolean;
    function GetMemoryFull: Boolean;
    procedure SetMaxMemorySize(NewValue: Longint);
    function GetFileEof: Boolean;
    function GetActualReadPos: LongInt;
    function GetActualLineReadPos: LongInt;
    function GetActualWritePos: LongInt;
  protected
  public
    constructor create(Const FileName: string; Mode: Word);
    destructor destroy; override;
    procedure FillMemory;
    function ReadLine: PChar;
    procedure WriteLine(NewLine: String);
    procedure FlushMemory;
    procedure Reset;
    property MaxMemorySize: Longint read fMaxMemorySize write SetMaxMemorySize;
    property Memory: PChar read fMemory write fMemory;
    property MemoryFull: Boolean read GetMemoryFull;
    property FileEof: Boolean read GetFileEof;
    property Eof: Boolean read fEof write fEof;
    property ActualReadPos: LongInt read GetActualReadPos;
    property ActualLineReadPos: LongInt read GetActualLineReadPos;
    property ActualWritePos: LongInt read GetActualWritePos;
  published
  end; { TLineStream }

implementation

constructor TLineStream.create(Const FileName: string; Mode: Word);
var
  fHandle: Integer;
begin
  if not FileExists(FileName) then
  begin
    fHandle := FileCreate(FileName);
    FileClose(fHandle);
  end;
  inherited create(FileName, Mode);
  fEof := False;
  MaxMemorySize := 32383;
  fMemorySize := 0;
  fMemoryPos := 0;
  Position := 0;
end; { create }

destructor TLineStream.destroy;
begin
  ReallocMem(fMemory, 0);
  inherited destroy;
end; { destroy }

function TLineStream.GetFileEof: Boolean;
begin
  if Position = Size then Result := True else Result := False;
end; { GetFileEof }

procedure TLineStream.SetMaxMemorySize(NewValue: Longint);
begin
  fMaxMemorySize := NewValue;
  ReallocMem(fMemory, fMaxMemorySize + 1);
end; { SetMaxMemorySize }

procedure TLineStream.FillMemory;
var
  Readed: LongInt;
begin
  Readed := Read(fMemory^, fMaxMemorySize);
  fMemorySize := Readed;
  if not FileEof then
  begin
    while (fMemory[fMemorySize - 1] in [#10, #13]) and (fMemorySize > 1) do dec(fMemorySize);
    while (not (fMemory[fMemorySize - 1] in [#10, #13])) and (fMemorySize > 1) do dec(fMemorySize);
  end;
  fMemory[fMemorySize] := #0;
  Position := Position - Readed + fMemorySize;
  fLineStart := 0;
end; { FillMemory }

function TLineStream.GetMemoryFull: Boolean;
begin
  if fMemorySize > 0 then Result := True else Result := False;
end; { GetMemoryFull }

function TLineStream.ReadLine: PChar;
var
  LineEnd: LongInt;
begin
  if fMemoryPos = fMemorySize then FillMemory;
  fMemoryPos := fLineStart;
  while fMemoryPos < fMemorySize do
    Case fMemory[fMemoryPos] of
      #13:
        begin
          if fMemory[fMemoryPos + 1] = #10 then
          begin
            LineEnd := fMemoryPos;
            inc(fMemoryPos, 2);
            fMemory[LineEnd] := #0;
            break;
          end else
          begin
            LineEnd := fMemoryPos;
            inc(fMemoryPos);
            fMemory[LineEnd] := #0;
            break;
          end;
        end;
      #10:
        begin
          LineEnd := fMemoryPos;
          inc(fMemoryPos);
          fMemory[LineEnd] := #0;
          break;
        end;
    else inc(fMemoryPos);
    end;
  Result := (fMemory + fLineStart);
  fLineStart := fMemoryPos;
  if (fMemoryPos = fMemorySize) and FileEof then Eof := True;
end; { ReadLine }

function TLineStream.GetActualReadPos: LongInt;
begin
  Result := Position - fMemorySize + fMemoryPos;
end; { GetActualReadPos }

function TLineStream.GetActualLineReadPos: LongInt;
begin
  Result := Position - fMemorySize + fLineStart;
end; { GetActualLineReadPos }

procedure TLineStream.WriteLine(NewLine: String);
var
  Count, Pos: Longint;
begin
  NewLine := NewLine + #13#10;
  Count := Length(NewLine);
  if (fMemoryPos >= 0) and (Count >= 0) then
  begin
    Pos := fMemoryPos + Count;
    if Pos > 0 then
    begin
      if Pos > FMaxMemorySize then
      begin
        FlushMemory;
      end;
      StrECopy((fMemory + fMemoryPos), PChar(NewLine));
      fMemoryPos := fMemoryPos + Count;
      fMemory[fMemoryPos] := #0;
    end;
  end;
end; { WriteLine }

function TLineStream.GetActualWritePos: LongInt;
begin
  Result := Position + fMemoryPos;
end; { GetActualWritePos }

procedure TLineStream.FlushMemory;
begin
  Write(fMemory^, fMemoryPos);
  fMemoryPos := 0;
end; { FlushMemory }

procedure TLineStream.Reset;
begin
  fEof := False;
  fMemorySize := 0;
  fMemoryPos := 0;
  Position := 0;
end; { Reset }

end.

